//      makeDB.c

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#define MAX_EPICS_CHANNEL	28	/* maximum EPICS channel name */
#define MAX_EPICS_DESC		28	/* maximum EPICS channel name */
#define MAX_EPICS_UNIT		15	/* maximum EPICS channel name */

#define MAX_CFILE_LENGTH 16 * 1024 * 1024
#define MAX_NAME_LENGTH 100
#define MAX_LIST_LENGTH 100
#define MAX_DB_FIELDS	100

#define OPC_PROP_CDT        1	/* canonocal data type */
#define OPC_PROP_VALUE      2	/* value */
#define OPC_PROP_QUALITY    3	/* data quality flag */
#define OPC_PROP_TIME       4	/* timestamp */
#define OPC_PROP_RIGHTS     5	/* access right: 1 read, 2 write, 3 read/write */
#define OPC_PROP_SCANRATE   6	/* scan rate */

#define OPC_PROP_UNIT       100	/* unit string -> EGU */
#define OPC_PROP_DESC       101	/* description string -> DESC */
#define OPC_PROP_HIEU       102	/* high expectation value -> HOPR */
#define OPC_PROP_LOEU       103	/* low expectation value -> LOPR */
#define OPC_PROP_HIRANGE    104	/* absolute maximum value -> DRVH */
#define OPC_PROP_LORANGE    105	/* absolute minumum value -> DRVL */
#define OPC_PROP_CLOSE      106	/* label for close state -> ONAM */
#define OPC_PROP_OPEN       107	/* label for open state -> ZNAM */
#define OPC_PROP_TIMEZONE   108	/* time zone */

#define OPC_PROP_FGC        201	/* foreground color */
#define OPC_PROP_BGC        202	/* background color */
#define OPC_PROP_BLINK      203	/* blinking */
#define OPC_PROP_BMP        204	/* bmp file */
#define OPC_PROP_SND        205	/* sound file */
#define OPC_PROP_HTML       206	/* html file */
#define OPC_PROP_AVI        207	/* avi file */

#define OPC_PROP_ALMSTAT    300
#define OPC_PROP_ALMHELP    301
#define OPC_PROP_ALMAREAS   302
#define OPC_PROP_ALMPRIMARYAREA  303
#define OPC_PROP_ALMCONDITION    304
#define OPC_PROP_ALMLIMIT   305
#define OPC_PROP_ALMDB      306
#define OPC_PROP_ALMHH      307	/* high high alaram -> HIHI */
#define OPC_PROP_ALMH       308	/* high alaram -> HIGH */
#define OPC_PROP_ALML       309	/* low alaram -> LOW */
#define OPC_PROP_ALMLL      310	/* low low alaram -> LOLO */
#define OPC_PROP_ALMROC     311
#define OPC_PROP_ALMDEV     312

#define OPT_OPC_SERVER      "_DEFOPC"	/* opc server option */
#define OPT_DEF_TSE			"_DEFTSE"	/* time stamp option */
#define OPT_ALL_OPC			"_ALLOPC"	/* automatic db generation option */

#define EPICS_DB_RECORD     "NAME"	/* channel name */
#define EPICS_DB_TYPE	    "TYPE"	/* channel type */
#define EPICS_DB_READONLY   "READONLY"	/* readonly flag */
#define EPICS_DB_OPC	    "OPC"	/* opc channel name */
#define EPICS_DB_DTYP	    "DTYP"	/* data type */
#define EPICS_DB_TSE	    "TSE"	/* time stamp */
#define EPICS_DB_PINI	    "PINI"	/* initialization */

#define EPICS_DB_EGU        "EGU"	/* unit string */
#define EPICS_DB_DESC       "DESC"	/* description string */
#define EPICS_DB_HOPR       "HOPR"	/* high ops value string */
#define EPICS_DB_LOPR       "LOPR"	/* low ops value string */
#define EPICS_DB_DRVH       "DRVH"	/* drive high string */
#define EPICS_DB_DRVL       "DRVL"	/* drive low string */
#define EPICS_DB_ONAM       "ONAM"	/* one name string */
#define EPICS_DB_ZNAM       "ZNAM"	/* zero name string */

#define EPICS_DB_FORBIDDEN	{"", "SCAN", "DOL", "INP", "VAL", "RVAL", "INIT", NULL}

#define EPICS_DB_ALLOWED	{EPICS_DB_RECORD, EPICS_DB_TYPE, EPICS_DB_READONLY, EPICS_DB_OPC, EPICS_DB_DTYP, EPICS_DB_TSE, EPICS_DB_PINI, \
							 EPICS_DB_EGU, EPICS_DB_DESC, EPICS_DB_HOPR, EPICS_DB_LOPR, EPICS_DB_DRVH, EPICS_DB_DRVL, \
							 EPICS_DB_ONAM, EPICS_DB_ZNAM, "LINR", "EGUF", "EGUL", "AOFF", "ASLO", "ESLO", "SMOO", \
							 "PREC", "HIHI", "LOLO", "HIGH", "LOW", "HHSV", "LLSV", "HSV", "LSV", "HYST", "ZSV", "OSV", "COSV", "IVOA", "IVOV", \
							 "ZRVL", "ONVL", "TWVL", "THVL", "FRVL", "FVVL", "SXVL", "SVVL", "EIVL", "NIVL", "TEVL", "ELVL", \
							 "TVVL", "TTVL", "FTVL", "FFVL", "ZRST", "ONST", "TWST", "THST", "FRST", "FVST", "SXST", "SVST", \
							 "EIST", "NIST", "TEST", "ELST", "TVST", "TTST", "FTST", "FFST", "UNSV",  "ONSV", "TWSV", "THSV", \
							 "FRSV", "FVSV", "SXSV", "SVSV", "EISV", "NISV", "TESV", "ELSV", "TVSV", "TTSV", "FTSV", "FFSV", NULL}

#define EPICS_DB_NUMVAL		{EPICS_DB_TSE, EPICS_DB_HOPR, EPICS_DB_LOPR, EPICS_DB_DRVH, EPICS_DB_DRVL, "EGUF", "EGUL", \
							 "AOFF", "ASLO", "ESLO", "SMOO", "PREC", "HIHI", "LOLO", "HIGH", "LOW", "HYST", \
                             "ZRVL", "ONVL", "TWVL", "THVL", "FRVL", "FVVL", "SXVL", "SVVL", "EIVL", "NIVL", "TEVL", "ELVL", \
							 "TVVL", "TTVL", "FTVL", "FFVL",}

const char* const epicsDisallowed[] = EPICS_DB_FORBIDDEN;
const char* const epicsNoComplain[] = EPICS_DB_ALLOWED;
const char* const epicsNumerical[] = EPICS_DB_NUMVAL;

/* type for global options */
struct global_opt_struct {
	int		allopc;
	int		deftse;
	char	defopc[MAX_NAME_LENGTH+10];
};

typedef struct global_opt_struct global_opt_type;


// check numerical
int checkNumerical (const char* num)
{
	double	x;
	char*	p;

	x = strtod (num, &p);
	while (p && isspace (*p)) ++p;
	return (p && (*p == '\0'));
}

// check validity of EPICS name //
int checkVarName (const char* opcname, char* epicsname)
{
	const char*	p;
	char*	q;

	if (!epicsname || !opcname) {
		return 0;
	}

	// do we need to convert?
	if (!*epicsname) {
		// ignore leading underscore
		for (; *opcname == '_'; ++opcname) ;
		// add ':' and change '_' to "H1PSL_", so we get "H1:PSL-"
		if ((strlen (opcname) > 6) && epicsname) {
			strncpy (epicsname, opcname, 2);
			epicsname[2] = ':';
			strncpy (epicsname + 3, opcname + 2, MAX_NAME_LENGTH - 3);
			epicsname[MAX_NAME_LENGTH] = '\0';
			if (epicsname[6] == '_') epicsname[6] = '-';
		}
		// go upper case
		for (q = epicsname; q && *q; ++q) *q = toupper (*q);
	}

	// make sure we have something of the form H1:PSL-
	if ((strlen (epicsname) < 6) || 
		!isalpha (epicsname[0]) || !isdigit (epicsname[1]) || (epicsname[2] != ':') ||
		!isalnum (epicsname[3]) || !isalnum (epicsname[4]) || !isalnum (epicsname[5]) || 
		(epicsname[6] != '-')) {
		fprintf (stderr, "WARNING: Unrecognized site/ifo/subsystem prefix in EPICS channel name %s.\n", opcname);
		//return 0;
	}
	// make sure channel is not too long
	if  (strlen (opcname) > MAX_EPICS_CHANNEL) {
		fprintf (stderr, "WARNING: EPICS channel name (%s) too long by %i.\n", opcname,
			     strlen (opcname) - MAX_EPICS_CHANNEL);
		return 0;
	}
	// check for special characters
	for (p = opcname + 6; p < opcname + strlen (opcname); ++p) {
		if (!isalnum (*p) && (*p != '_')) {
			fprintf (stderr, "WARNING: Illegale character in EPICS channel name %s.\n", opcname);
			return 0;
		}
	}

	return 1;
}

// check validity of type name //
int convertVarType (const char* opctype, char* epicstype)
{
	char*	p;
	char	t[MAX_NAME_LENGTH+10];

	if (!opctype) {
		if (epicstype) *epicstype = '\0'; 
		return 0;
	}
	strncpy (t, opctype, MAX_NAME_LENGTH);
	t[MAX_NAME_LENGTH] = '\0';
	for (p = t; p && *p; ++p) *p = toupper (*p);
	if (strcmp (t, "BOOL") == 0) {
		if (epicstype) strcpy (epicstype, "bo");
		return 1;
	}
	else if ((strcmp (t, "SINT") == 0) || (strcmp (t, "INT") == 0) ||
		(strcmp (t, "DINT") == 0) || (strcmp (t, "LINT") == 0) ||
		(strcmp (t, "USINT") == 0) || (strcmp (t, "UINT") == 0) ||
		(strcmp (t, "UDINT") == 0) || (strcmp (t, "ULINT") == 0) ||
		(strcmp (t, "BYTE") == 0) || (strcmp (t, "WORD") == 0) ||
		(strcmp (t, "DWORD") == 0) || (strcmp (t, "LWORD") == 0)) {
		if (epicstype) strcpy (epicstype, "longout");
		return 1;
	}
	else if ((strcmp (t, "REAL") == 0) || (strcmp (t, "LREAL") == 0)) {
		if (epicstype) strcpy (epicstype, "ao");
		return 1;
	}
	else if (strcmp (t, "STRING") == 0) {
		if (epicstype) strcpy (epicstype, "stringout");
		return 1;
	}
	else {
		/*fprintf (stderr, "WARNING: Unrecognized OPC data type %s.\n", t);*/
		if (epicstype) strcpy (epicstype, "ao");
		return 0;
	}
	return 0;
}

// find EPICS database field
char* findEpicsDb (char** list, const char* name)
{
	int		i;
	char*	p;
	char	t[MAX_NAME_LENGTH+10];

	if (!list || !name) {
		return NULL;
	}
	strncpy (t, name, MAX_NAME_LENGTH);
	t[MAX_NAME_LENGTH] = '\0';
	for (p = t; *p; ++p) {
		*p = toupper (*p);
	}

	for (i = 0; i < MAX_DB_FIELDS; ++i) {
		if (!list[i]) {
			return NULL;
		}
		if (strcmp (list[i], t) == 0) {
			return list[i];
		}
	}
	return NULL;
}

// insert EPICS database field
char* insertEpicsDb (char** list, const char* name, const char* arg)
{
	char*	p = NULL;
	char*	q = NULL;
	const char*	r;
	int		i;
	char	t[MAX_NAME_LENGTH+10];

	if (!list || !name) {
		return NULL;
	}
	strncpy (t, name, MAX_NAME_LENGTH);
	t[MAX_NAME_LENGTH] = '\0';
	for (p = t; *p; ++p) {
		*p = toupper (*p);
	}
	for (i = 0; epicsDisallowed[i]; ++i) {
		if (strcmp (t, epicsDisallowed[i]) == 0) {
			fprintf (stderr, "ERROR: Illegale EPICS database field %s.\n", t);
			return NULL;
		}
	}
	r = NULL;
	for (i = 0; epicsNoComplain[i]; ++i) {
		if (strcmp (t, epicsNoComplain[i]) == 0) {
			r = epicsNoComplain[i];
			break;
		}
	}
	if (!r) {
		fprintf (stderr, "WARNING: Unusual EPICS database field %s.\n", t);
	}

	p = findEpicsDb (list, t);

	if (!p) {
		for (i = 0; i < MAX_DB_FIELDS; ++i) {
			if (!list[i]) {
				list[i] = (char*)malloc (2 * MAX_NAME_LENGTH + 20);
				p = list[i];
				strncpy (p, t, MAX_NAME_LENGTH);
				p[MAX_NAME_LENGTH] = '\0';
				break;
			}
		}
	}
	if (!p) {
		return NULL;
	}
	q = p + strlen (p) + 1;
	if (arg) {
		strncpy (q, arg, MAX_NAME_LENGTH);
		q[MAX_NAME_LENGTH] = '\0';
	}
	else *q = '\0';
	return p;
}

// delete EPICS database fields
void deleteEpicsDb (char*** list)
{
	int		i;

	if (!list || !*list) return;
	for (i = 0; i < MAX_DB_FIELDS + 2; ++i) {
		if ((*list)[i]) free ((*list)[i]);
	}
	free (*list);
	*list = NULL;
}

// initialize EPICS database field
int initEpicsDb (char*** list, const char* varname, const char* typname)
{
	int		i;
	
	deleteEpicsDb (list);
	*list = (char**)malloc ((MAX_DB_FIELDS + 3) * sizeof (char*));
	if (*list == NULL) {
		return 0;
	}
	for (i = 0; i < MAX_DB_FIELDS + 3; ++i) {
		(*list)[i] = NULL;
	}
	if (!varname || !typname) {
		return 0;
	}
	insertEpicsDb (*list, EPICS_DB_RECORD, ""); /* index 0 */
	insertEpicsDb (*list, EPICS_DB_TYPE, typname);
	insertEpicsDb (*list, EPICS_DB_READONLY, "0");
	insertEpicsDb (*list, EPICS_DB_OPC, varname);
	insertEpicsDb (*list, EPICS_DB_DTYP, "opc"); /* index 4 */
	insertEpicsDb (*list, EPICS_DB_TSE, "-2");
	insertEpicsDb (*list, EPICS_DB_PINI, "0");

	return 1;
}

// initialize EPICS database field
int copyEpicsDb (char*** l1, char** l2)
{
	int		i;

	if (!l1) return 0;
	deleteEpicsDb (l1);
	if (!l2) {
		*l1 = NULL;
		return 0;
	}
	*l1 = (char**)malloc ((MAX_DB_FIELDS + 3) * sizeof (char*));
	if (*l1 == NULL) {
		return 0;
	}
	for (i = 0; i < MAX_DB_FIELDS + 3; ++i) {
		(*l1)[i] = NULL;
	}
	for (i = 0; (i < MAX_DB_FIELDS) && l2[i]; ++i) {
		(*l1)[i] = (char*)malloc (2 * MAX_NAME_LENGTH + 20);
		if (!(*l1)[i]) {
			return 0;
		}
		memcpy ((*l1)[i], l2[i], 2 * MAX_NAME_LENGTH + 20);
	}
	return 1;
}

// check EPICS database record
int checkEpicsDb (char** list)
{
	char*	p;
	char*	q;
	const char* const*	l;

	if (!list || !list[0] || !list[1] || !list[2] || !list[3] || !list[4]) {
		return 0;
	}

	/* makes sure name is ok */
	p = list[0] + strlen (list[0]) + 1;
	q = list[3] + strlen (list[3]) + 1;
	if (!checkVarName (q, p)) {
		return 0;
	}

	/* makes sure input/output is set correct */
	p = list[1] + strlen (list[1]) + 1;
	for (; *p; ++p) {
		*p = tolower (*p);
	}
	p = list[1] + strlen (list[1]) + 1;
	q = list[2] + strlen (list[2]) + 1;
	if (*q == '1') {
		if (strcmp (p, "stringout") == 0) {
			strcpy (p, "stringin");
		}
		else if (strcmp (p, "ao") == 0) {
			strcpy (p, "ai");
		}
		else if (strcmp (p, "longout") == 0) {
			strcpy (p, "longin");
		}
		else if (strcmp (p, "bo") == 0) {
			strcpy (p, "bi");
		}
		else if (strcmp (p, "mbbo") == 0) {
			strcpy (p, "mbbi");
		}
		else if (strcmp (p, "mbboDirect") == 0) {
			strcpy (p, "mbbiDirect");
		}
		else if ((strcmp (p, "stringin") != 0) &&
			(strcmp (p, "ai") != 0) &&
			(strcmp (p, "longin") != 0) &&
			(strcmp (p, "bi") != 0) &&
			(strcmp (p, "mbbi") != 0) &&
			(strcmp (p, "mbbiDirect") != 0)) {
			fprintf (stderr, "ERROR: Unknown or unsupported EPICS record type %s in %s.\n", 
				p, list[0] + strlen (list[0]) + 1);
			return 0;
		}
	}
	else {
		if (strcmp (p, "stringin") == 0) {
			strcpy (p, "stringout");
		}
		else if (strcmp (p, "ai") == 0) {
			strcpy (p, "ao");
		}
		else if (strcmp (p, "longin") == 0) {
			strcpy (p, "longout");
		}
		else if (strcmp (p, "bi") == 0) {
			strcpy (p, "bo");
		}
		else if (strcmp (p, "mbbi") == 0) {
			strcpy (p, "mbbo");
		}
		else if (strcmp (p, "mbbiDirect") == 0) {
			strcpy (p, "mbboDirect");
		}
		else if ((strcmp (p, "stringout") != 0) &&
			(strcmp (p, "ao") != 0) &&
			(strcmp (p, "longout") != 0) &&
			(strcmp (p, "bo") != 0) &&
			(strcmp (p, "mbbo") != 0) &&
			(strcmp (p, "mbboDirect") != 0)) {
			fprintf (stderr, "ERROR: Unknown or unsupported EPICS record type %s in %s.\n", 
				p, list[0] + strlen (list[0]) + 1);
			return 0;
		}
	}

	/* makse sure DTYP is ok */
	p = list[4] + strlen (list[4]) + 1;
	if ((strcmp (p, "opc") != 0) && (strcmp (p, "opcRaw") != 0)) {
		fprintf (stderr, "ERROR: Unknown opc access type (DTYP) %s in %s.\n", 
			p, list[0] + strlen (list[0]) + 1);
		return 0;
	}

	/* check length of description */
	p = findEpicsDb (list, EPICS_DB_DESC);
	if (p) {
		p = p + strlen (p) + 1;
		if (strlen (p) > MAX_EPICS_DESC) {
			fprintf (stderr, "ERROR: EPICS record description in %s too long (%s) by %i.\n", 
				list[0] + strlen (list[0]) + 1, p, strlen (p) - MAX_EPICS_DESC);
			return 0;
		}
	}
	/* check length of units string */
	p = findEpicsDb (list, EPICS_DB_EGU);
	if (p) {
		p = p + strlen (p) + 1;
		if (strlen (p) > MAX_EPICS_UNIT) {
			fprintf (stderr, "ERROR: EPICS record unit string in %s too long (%s) by %i.\n", 
				list[0] + strlen (list[0]) + 1, p, strlen (p) - MAX_EPICS_UNIT);
			return 0;
		}
	}
	/* check numerical values */
	for (l = epicsNumerical; l && *l; ++l) {
		p = findEpicsDb (list, *l);
		if (p) {
			q = p + strlen (p) + 1;
 			if (!checkNumerical (q)) {
				fprintf (stderr, "WARNING: EPICS field %s in %s should be numerical (%s).\n", 
					p, list[0] + strlen (list[0]) + 1, q);
			}
		}
	}

	return 1;
}

// write EPICS database record
int writeEpicsDb (FILE* db, char** list, global_opt_type* gopt)
{
	char*	p;
	char*	q;
	int		i;

	if (!list || !checkEpicsDb (list)) {
		return 0;
	}
	p = list[0] + strlen (list[0]) + 1;
	q = list[1] + strlen (list[1]) + 1;
	fprintf(db, "record(%s,\"%s\") {\n", q, p);
	p = list[2] + strlen (list[2]) + 1;
	q = list[3] + strlen (list[3]) + 1;
	if (*p != '1') {
		fprintf (db, "\tfield(SCAN,\"Passive\")\n");
		fprintf (db, "\tfield(DOL,\"@%s.%s\")\n", gopt ? gopt->defopc : "opc", q);
	}
	else {
		fprintf (db, "\tfield(SCAN,\"I/O Intr\")\n");
		fprintf (db, "\tfield(INP,\"@%s.%s\")\n", gopt ? gopt->defopc : "opc", q);
	}
	for (i = 4; i < MAX_DB_FIELDS; ++i) {
		p = list[i];
		if (p == NULL) break;
		q = p + strlen (p) + 1;
		fprintf (db, "\tfield(%s,\"%s\")\n", p, q);
	}
	fprintf(db, "}\n");
	return 1;
}

int globalOption (const char* epicsName, const char* epicsArg, global_opt_type* gopt)
{
	char	opt[MAX_NAME_LENGTH+10];
	const char* p;
	char* q;

	if (!epicsName || !epicsArg || !gopt) {
		return 0;
	}
	for (p = epicsName, q = opt; *p && (q - opt) < MAX_NAME_LENGTH; ++p, ++q) {
		*q = toupper (*p);
	}
	*q = '\0';

	if (strcmp (opt, OPT_OPC_SERVER) == 0) {
		strncpy (gopt->defopc, epicsArg, MAX_NAME_LENGTH);
		gopt->defopc[MAX_NAME_LENGTH] = '\0';
		return 1;
	}
	else if (strcmp (opt, OPT_DEF_TSE) == 0) {
		gopt->deftse = atoi (epicsArg);
		return 1;
	}
	else if (strcmp (opt, OPT_ALL_OPC) == 0) {
		gopt->allopc = (atoi (epicsArg) == 1);
		return 1;
	}
	return 0;
}


// main function for parsing the .txt file of the TwinCAT variables
void main (int argc, const char* argv[]) {
	
	/* variables pertaining to the parsing */
	char*	charPtr = NULL;
	char*	prevPtr = NULL;
	char*	varPtr = NULL;
	char*	typPtr = NULL;
	char*	p = NULL;
	char*	q = NULL;
	int		i;
	int		isComment = 0;
	char*	varName = NULL;
	char*	typName = NULL;
	int		isValid = 0; /* valid variable */
	int		isOPC = 0; /* OPC comment */
	int		isOPCDef = 0; /* OPC definition */
	char	opcName[MAX_NAME_LENGTH+10];
	char	opcArg[MAX_NAME_LENGTH+10];
	int		isOPCchn = 0; /* is this an OPC export/import channel? */
	int		isEPICSonOPC = 0; /* generate EPICS record from OPC */
	int		isEPICS = 0; /* EPICS comment */
	int		isEPICSDef = 0; /* EPICS definition */
	char	epicsName[MAX_NAME_LENGTH+10];
	char	epicsArg[MAX_NAME_LENGTH+10];
	int		isValidOPC = 0; /* valid OPC comment */
	int		isValidEPICS = 0; /* valid EPICS comment */
	char**	epicsDbBase = 0; /* Epics database as determined by OPC */
	char**	epicsDb = 0; /* Epics database */
	char*	firstDelimiter = NULL;
	char	dummy[MAX_NAME_LENGTH+10];
	int		DbCount = 0;

	/* variables pertaining to global options */
	global_opt_type	gopt = {0, -2, "opc"};

	/* variables pertaining to the files */
	int		fLength = 0;
	char*	cFile = NULL;
	FILE*	file = NULL;
	FILE*	DB = NULL;

	/* check arguments */
	if (argc < 2) {
		fprintf(stderr,"Please provide the path to the TwinCAT global variables file as an input argument.\n");
		exit (1);
	}

	/* load varFile */
	file = fopen(argv[1],"rt");
	if (file == NULL) {
		fprintf(stderr,"%s %s %s","ERROR: TwinCAT global variables file ", argv[1]," does not exist. Aborting.\n");
		exit(1);
	}
	fseek(file, 0L, SEEK_END);
	fLength = ftell(file);
	if (fLength > MAX_CFILE_LENGTH) {
		fprintf (stderr, "WARNING: Input file length %i larger than maximum allowed (%i). Truncated.\n", fLength, MAX_CFILE_LENGTH);
		fLength = MAX_CFILE_LENGTH;
	}
	rewind(file);
	cFile = malloc (fLength + 10);
	fread(cFile, 1, fLength, file);
	cFile[fLength] = '\0';
	fclose (file);
	
	/* make output file */
	DB = NULL;
	if (argc == 3) {
		DB = fopen(argv[2],"w");
		if (DB == NULL) {
			fprintf(stderr,"%s %s %s","ERROR: Output file", argv[2],"does not exist. Aborting.\n");
			exit(1);
		}
	}
	else if (argc < 3) {
		fprintf(stderr,"WARNING: No output file specified, writing to stdout\n");
		DB = fopen("stdout","w");
		if (DB == NULL) {
			fprintf(stderr,"%s %s %s","ERROR: Output file", "stderr", "does not exist. Aborting.\n");
			exit(1);
		}
	}
	
	/* read each token in file */
	charPtr = strtok (cFile, " \r\n\t");
	prevPtr = NULL;
	varPtr = NULL;
	typPtr = NULL;
	varName = NULL;
	typName = NULL;
	isComment = 0;
	isOPC =  0;
	isOPCDef =  0;
	isEPICS = 0;
	isEPICSDef = 0;
	isValid = 0;
	isValidOPC = 0;
	isValidEPICS = 0;
	isEPICSonOPC = 0;
	DbCount = 0;
	while (1) {
		/* check for automatic EPICS record generation 
		   also make sure to get the last one! */
		if ((!charPtr && isEPICSonOPC) ||
			(charPtr && !isComment && isEPICSonOPC && !(charPtr[0] == '\0') && 
			!((charPtr[0] == '(') && (charPtr[1] == '*')))) {
			// anything but a comment
			/* make new EPICS database from OPC template */
			copyEpicsDb (&epicsDb, epicsDbBase);
			sprintf (dummy, "%d", gopt.deftse);
			insertEpicsDb (epicsDb, EPICS_DB_TSE, dummy);
			/* convert opc data type into EPICS */
			if (epicsDb && epicsDb[1]) {
				p = epicsDb[1] + strlen (epicsDb[1]) + 1;
				convertVarType (p, p);
			}
			epicsName[0] = '\0';
			isValidEPICS = (epicsDb && epicsDbBase);
			/* write an EPICS record if valid */
			if (isValid && isValidOPC && isValidEPICS) {
				if (!writeEpicsDb (DB, epicsDb, &gopt)) {
					if (varName) fprintf (stderr, "ERROR: Unable to write EPICS dB record for (%s).\n", varName);
				}
				else ++DbCount;
			}
			isValidEPICS = 0;
			isEPICSonOPC = 0;
		}

		/* exit at end of input */
		if (!charPtr) break;

		/* check for var type: either before assignment or before semi-colon
		   yes, the buffer is long enough! */
		if (!isComment && varPtr && !typPtr && (charPtr[0] == ':') && (charPtr[1] == '=')) {
			typPtr = prevPtr;
			isValid = 0;
			*charPtr++ = '\0';
			prevPtr = charPtr;
			++charPtr;
			continue;
		}
		/* check for end of declaration */
		if (!isComment && (charPtr[0] == ';')) {
			*charPtr++ = '\0';
			varName = varPtr;
			if (typPtr) {
				typName = typPtr;
			}
			else {
				typName = prevPtr;
			}
			/* only use upper case for EPICS */
			if (typName) {
				/* only use upper case for EPICS */
				for (p = typName; p && *p; ++p) *p = toupper (*p);
			}

			/* check validity */
			isValid = varName && typName;
			if (!isValid) {
				varName = NULL;
				typName = NULL;
			}
			else {
				initEpicsDb (&epicsDbBase, varName, typName);
				deleteEpicsDb (&epicsDb);
			}
			isOPCchn = 0;
			isValidOPC = 0;
			isValidEPICS = 0;

			/* reset var/type pointer after sem-column separator */
			varPtr = NULL;
			typPtr = NULL;
			prevPtr = charPtr;
			continue;
		}
		/* check for variables: either before column or before AT */
		if (!isComment && !varPtr && (charPtr[0] == ':')) {
			varPtr = prevPtr;
			varName = NULL;
			isValid = 0;
			*charPtr++ = '\0';
			prevPtr = charPtr;
			continue;
		}
		if (!isComment && !varPtr && (charPtr[0] == 'A') && (charPtr[1] == 'T')) {
			varPtr = prevPtr;
			varName = NULL;
			isValid = 0;
			prevPtr = charPtr++;
			++charPtr;
			continue;
		}
		/* check for OPC comment */
		if (!isComment && (charPtr[0] == '(') && (charPtr[1] == '*') && (charPtr[2] == '~')) {
			isComment = 1;
			isOPC = 1;
			isOPCDef = 0;
			isValidOPC = 1;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 3;
			continue;
		}
		/* check for EPICS comment */
		if (!isComment && (charPtr[0] == '(') && (charPtr[1] == '*') && (charPtr[2] == '&')) {
			if (isEPICSonOPC) isEPICSonOPC = 0; // no longer needed
			/* make new EPICS database from OPC template */
			copyEpicsDb (&epicsDb, epicsDbBase);
			sprintf (dummy, "%d", gopt.deftse);
			insertEpicsDb (epicsDb, EPICS_DB_TSE, dummy);
			/* convert opc data type into EPICS */
			if (epicsDb && epicsDb[1]) {
				p = epicsDb[1] + strlen (epicsDb[1]) + 1;
				convertVarType (p, p);
			}
			epicsName[0] = '\0';

			isComment = 1;
			isOPC = 0;
			isEPICS = 1;
			isEPICSDef = 0;
			isValidEPICS = (epicsDb && epicsDbBase);
			prevPtr = charPtr;
			charPtr = charPtr + 3;
			continue;
		}
		/* check for normal comment */
		if (!isComment && (charPtr[0] == '(') && (charPtr[1] == '*')) {
			isComment = 1;
			isOPC = 0;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 2;
			continue;
		}
		/* check for end of OPC comment */
		if (isComment && isOPC && (charPtr[0] == '*') && (charPtr[1] == ')')) {
			if (gopt.defopc) isEPICSonOPC = 1;
			else isEPICSonOPC = 0;
			/* reset comment */
			isComment = 0;
			isOPC = 0;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 2;
			continue;
		}
		/* check for end of EPICS comment with silent option */
		if (isComment && isEPICS && (charPtr[0] == '*') && (charPtr[1] == '*') && (charPtr[2] == ')')) {
			/* reset comment */
			isComment = 0;
			isOPC = 0;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 3;
			continue;
		}
		/* check for end of EPICS comment */
		if (isComment && isEPICS && (charPtr[0] == '*') && (charPtr[1] == ')')) {
			/* write an EPICS record if valid */
			if (isValid && isValidOPC && isValidEPICS) {
				if (!writeEpicsDb (DB, epicsDb, &gopt)) {
					if (varName) fprintf (stderr, "ERROR: Unable to write EPICS dB record for (%s).\n", varName);
				}
				else ++DbCount;
			}
			isValidEPICS = 0;

			/* reset comment */
			isComment = 0;
			isOPC = 0;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 2;
			continue;
		}
		/* check for end of any other comment */
		if (isComment && (charPtr[0] == '*') && (charPtr[1] == ')')) {
			/* reset comment */
			isComment = 0;
			isOPC = 0;
			isEPICS = 0;
			prevPtr = charPtr;
			charPtr = charPtr + 2;
			continue;
		}

		/* decipher OPC comments */
		if (isComment && isOPC && !isOPCDef && (charPtr[0] == '(')) {
			isOPCDef = 1;
			*opcName ='\0';
			*opcArg = '\0';
			prevPtr = charPtr++;
			continue;
		}
		if (isComment && isOPC && isOPCDef && (charPtr[0] == ')')) {
			isOPCDef = 0;
			prevPtr = charPtr++;
			continue;
		}
		if (isComment && isOPC && isOPCDef && (charPtr[0] == ':')) {
			*charPtr++ = '\0';
			if (++isOPCDef == 3) {
				if (strcmp (opcName, "OPC") == 0) {
					isOPCchn = (atoi (prevPtr) != 0);
				}
				else if ((strncmp (opcName, "OPC_PROP", 6) == 0) &&
					(p = strchr (opcName, '['))) {
					switch (atoi (++p)) {
						case OPC_PROP_RIGHTS:
							i = atoi (opcArg);
							if (i == 2) {
								i = 3;
								if (varName) fprintf (stderr, "WARNING: EPICS doesn't support write-only (%s). Use R/W instead.\n", varName);
							}
							if ((i < 1) || (i > 3)) {
								i = 3;
								if (varName) fprintf (stderr, "WARNING: Unrecognized OPC access right %s (%s). Use R/W instead.\n", opcArg, varName);
							}
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_READONLY, i == 1 ? "1" : "0")) {
								if (varName) fprintf (stderr, "ERROR: Unable to add in/out type for %s.\n", varName);
							}
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_PINI, i == 1 ? "1" : "0")) {
								if (varName) fprintf (stderr, "ERROR: Unable to add PINI for %s.\n", varName);
							}
							break;
						case OPC_PROP_UNIT:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_EGU, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_EGU, opcArg, varName);
							}
							break;
						case OPC_PROP_DESC:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_DESC, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_DESC, opcArg, varName);
							}
							break;
						case OPC_PROP_HIEU:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_HOPR, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_HOPR, opcArg, varName);
							}
							break;
						case OPC_PROP_LOEU:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_LOPR, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_LOPR, opcArg, varName);
							}
							break;
						case OPC_PROP_HIRANGE:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_DRVH, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_DRVH, opcArg, varName);
							}
							break;
						case OPC_PROP_LORANGE:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_DRVL, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_DRVL, opcArg, varName);
							}
							break;
						case OPC_PROP_CLOSE:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_ONAM, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_ONAM, opcArg, varName);
							}
							break;
						case OPC_PROP_OPEN:
							if (!insertEpicsDb (epicsDbBase, EPICS_DB_ZNAM, opcArg)) {
								if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", EPICS_DB_ZNAM, opcArg, varName);
							}
							break;
						default:
							if (varName) fprintf (stderr, "WARNING: Unrecognized OPC property %s for %s.\n", opcName, varName);
							break;
					}
				}
				else {
					if (varName) fprintf (stderr, "WARNING: Unrecognized OPC command %s for %s.\n", opcName, varName);
				}
			}
			prevPtr = charPtr;
			continue;
		}
		if (isComment && isOPC && (isOPCDef == 1)) {
			for (p = charPtr, q = opcName + strlen (opcName); 
				 *p && (*p != ':') && (q - opcName < MAX_NAME_LENGTH); ++p) {
				*q++ = toupper (*p);
				*q = '\0';
			}
			prevPtr = charPtr;
		}
		if (isComment && isOPC && (isOPCDef == 2)) {
			q = opcArg + strlen (opcArg);
			if ((strlen (charPtr) > 0) && (strlen (opcArg) < MAX_NAME_LENGTH) &&
				(strlen (opcArg) > 0)) {
				*q++ = ' ';
			}
			for (p = charPtr; *p && (*p != ':') && (q - opcArg < MAX_NAME_LENGTH); ++p, ++q) {
				*q = *p;
			}
			*q = '\0';
			prevPtr = charPtr;
		}

		/* decipher EPICS comment */
		if (isComment && isEPICS && (isEPICSDef == 0) && (charPtr[0] == ':')) {
			if (strlen (epicsName) > 0) {
				epicsArg[0] = '\0';
				isEPICSDef = 1;
			}
			charPtr++;
			continue;
		}
		if (isComment && isEPICS && (isEPICSDef == 0)) {
			for (p = charPtr, q = epicsName; 
				 *p && (*p != ':') && (q - epicsName < MAX_NAME_LENGTH); ++p) {
				*q++ = *p;
			}
			*q = '\0';
			prevPtr = charPtr;
			if (*p == ':') {
				charPtr = p;
				continue;
			}
		}
		if (isComment && isEPICS && (isEPICSDef >= 1) && charPtr[0]) {
			if ((isEPICSDef == 1) && (charPtr[0] == '"')) {
				isEPICSDef = 2;
				charPtr++;
				continue;
			}
			q = epicsArg + strlen (epicsArg);
			if ((strlen (charPtr) > 0) && (strlen (epicsArg) < MAX_NAME_LENGTH) &&
				(strlen (epicsArg) > 0)) {
				*q++ = ' ';
			}
			for (p = charPtr; *p && (*p != '"') && ((isEPICSDef != 1) || (*p != '*')) && 
					(q - epicsArg < MAX_NAME_LENGTH); ++p) {
				*q++ = *p;
			}
			*q = '\0';
			if ((isEPICSDef == 2) && (*p == '"')) {
				isEPICSDef = 1;
			}
			if (isEPICSDef == 1) {
				isEPICSDef = 0;
			}
			if (isEPICSDef == 0) {
				if (globalOption (epicsName, epicsArg, &gopt)) {
				}
				else if (!insertEpicsDb (epicsDb, epicsName, epicsArg)) {
					if (varName) fprintf (stderr, "ERROR: Unable to add %s:%s for %s.\n", epicsName, epicsArg, varName);
				}
			}
			prevPtr = charPtr;
			if (*p == '"') {
				charPtr = p + 1;
				continue;
			}
		}

		/* check for paranethesis, colon and semi-colon further into the string
		   make sure we are getting the first occurance */
		firstDelimiter = NULL;
		if ((strlen (charPtr) > 0 ) && (p = strchr (charPtr + 1, ':'))) {
			firstDelimiter = p;
		}
		if ((strlen (charPtr) > 0 ) && (p = strchr (charPtr + 1, ';'))) {
			if (!firstDelimiter || (p < firstDelimiter)) firstDelimiter = p;
		}
		if (!isComment && (strlen (charPtr) > 0 ) && (p = strchr (charPtr + 1, '('))) {
			if (!firstDelimiter || (p < firstDelimiter)) firstDelimiter = p;
		}
		if (isComment && (strlen (charPtr) > 0 ) && (p = strchr (charPtr + 1, '*'))) {
			if (!firstDelimiter || (p < firstDelimiter)) firstDelimiter = p;
		}
		if (isComment && isOPC && (strlen (charPtr) > 0 ) && (p = strchr (charPtr + 1, ')'))) {
			if (!firstDelimiter || (p < firstDelimiter)) firstDelimiter = p;
		}
		if (firstDelimiter) {
			prevPtr = charPtr;
			charPtr = firstDelimiter;
			continue;
		}

		/* move on to next token */
		if (strlen (charPtr) > 0) prevPtr = charPtr;
		charPtr = strtok (NULL, " \r\n\t");
	}
	/* save file */
	fclose(DB);
	free (cFile);
	/* final count */
	fprintf (stderr, "Generated %i EPICS datbase records.\n", DbCount);
}
